#include <gtk/gtklabel.h>
#include <gtk/gtkpixmap.h>
#include <gtk/gtktable.h>
#include <gtk/gtktree.h>
#include "gtktreeitempix.h"

static GtkTreeItemClass *parent_class = NULL;

static void gtk_tree_item_pix_class_init     (GtkTreeItemPixClass *klass);
static void gtk_tree_item_pix_init           (GtkTreeItemPix *tree_item_pix);
static void gtk_tree_item_pix_realize        (GtkWidget *widget);
static void gtk_tree_item_pix_destroy        (GtkObject *object);

static void tree_item_pix_collapse           (GtkWidget *item, gpointer data);
static void tree_item_pix_expand             (GtkWidget *item, gpointer data);

static void gtk_tree_item_pix_add_pixmaps    (GtkTreeItemPix *tree_item);
static void gtk_tree_item_pix_remove_pixmaps (GtkTreeItemPix *tree_item);

GtkType
gtk_tree_item_pix_get_type (void)
{

  static GtkType tree_item_pix_type = 0;

  if (!tree_item_pix_type)
    {
      GtkTypeInfo tree_item_pix_info =
      {
        "GtkTreeItemPix",
        sizeof (GtkTreeItemPix),
        sizeof (GtkTreeItemPixClass),
        (GtkClassInitFunc) gtk_tree_item_pix_class_init,
        (GtkObjectInitFunc) gtk_tree_item_pix_init,
        /* reserved_1 */ NULL,
        /* reserved_2 */ NULL,
        (GtkClassInitFunc) NULL,
      };

      tree_item_pix_type = gtk_type_unique (gtk_tree_item_get_type (),
                                            &tree_item_pix_info);
    }

  return tree_item_pix_type;
}

static void
gtk_tree_item_pix_class_init (GtkTreeItemPixClass *class)
{
  GtkObjectClass *object_class;
  GtkWidgetClass *widget_class;

  object_class = (GtkObjectClass*) class;
  widget_class = (GtkWidgetClass*) class;

  parent_class = gtk_type_class (gtk_tree_item_get_type ());

  object_class->destroy = gtk_tree_item_pix_destroy;

  widget_class->realize = gtk_tree_item_pix_realize;
}

static void
gtk_tree_item_pix_init (GtkTreeItemPix *tree_item_pix)
{
  GtkWidget *pixmapwid, *tablewid, *labelwid;

  g_return_if_fail (tree_item_pix != NULL);
  g_return_if_fail (GTK_IS_TREE_ITEM_PIX (tree_item_pix));

  tree_item_pix->label_position = GTK_TREE_ITEM_PIX_LABEL_LAST;

  tablewid = gtk_table_new (1, 2, FALSE);
  gtk_widget_show (tablewid);
  gtk_widget_ref (tablewid);
  gtk_container_add (GTK_CONTAINER (tree_item_pix), tablewid);
  tree_item_pix->table_widget = tablewid;

  labelwid = gtk_label_new(" ");
  gtk_misc_set_alignment (GTK_MISC (labelwid), 0.0, 0.5);
  gtk_widget_show (labelwid);
  gtk_widget_ref (labelwid);
  if (tree_item_pix->label_position == GTK_TREE_ITEM_PIX_LABEL_LAST)
    gtk_table_attach (GTK_TABLE (tablewid), labelwid, 1, 2, 0, 1,
      GTK_SHRINK, GTK_SHRINK, 0, 0);
  else
    gtk_table_attach (GTK_TABLE (tablewid), labelwid, 0, 1, 0, 1,
      GTK_SHRINK, GTK_SHRINK, 0, 0);
  tree_item_pix->label_widget = labelwid;

  pixmapwid = gtk_type_new (gtk_pixmap_get_type ());
  gtk_widget_show (pixmapwid);
  if (!GTK_TREE_ITEM (tree_item_pix)->expanded)
  {
    if (tree_item_pix->label_position == GTK_TREE_ITEM_PIX_LABEL_LAST)
      gtk_table_attach (GTK_TABLE (tablewid), pixmapwid, 0, 1, 0, 1,
        GTK_SHRINK, GTK_SHRINK, 0, 0);
    else
      gtk_table_attach (GTK_TABLE (tablewid), pixmapwid, 1, 2, 0, 1,
        GTK_SHRINK, GTK_SHRINK, 0, 0);
  }
  tree_item_pix->collapsed_pix_widget = pixmapwid;
  gtk_widget_ref (tree_item_pix->collapsed_pix_widget);
  gtk_object_sink (GTK_OBJECT (tree_item_pix->collapsed_pix_widget));

  pixmapwid = gtk_type_new (gtk_pixmap_get_type ());
  gtk_widget_show (pixmapwid);
  if (GTK_TREE_ITEM (tree_item_pix)->expanded)
  {
    if (tree_item_pix->label_position == GTK_TREE_ITEM_PIX_LABEL_LAST)
      gtk_table_attach (GTK_TABLE (tablewid), pixmapwid, 0, 1, 0, 1,
        GTK_SHRINK, GTK_SHRINK, 0, 0);
    else
      gtk_table_attach (GTK_TABLE (tablewid), pixmapwid, 1, 2, 0, 1,
        GTK_SHRINK, GTK_SHRINK, 0, 0);
  }
  tree_item_pix->expanded_pix_widget = pixmapwid;
  gtk_widget_ref (tree_item_pix->expanded_pix_widget);
  gtk_object_sink (GTK_OBJECT (tree_item_pix->expanded_pix_widget));

  gtk_signal_connect (GTK_OBJECT (tree_item_pix), "collapse",
                      GTK_SIGNAL_FUNC (tree_item_pix_collapse), NULL);
  gtk_signal_connect (GTK_OBJECT (tree_item_pix), "expand",
                      GTK_SIGNAL_FUNC (tree_item_pix_expand), NULL);
}

GtkWidget*
gtk_tree_item_pix_new (gchar *label,
                       gchar **collapsed_xpm,
                       gchar **expanded_xpm,
                       guint location)
{
  GtkWidget *tree_item_pix;

  tree_item_pix = GTK_WIDGET (gtk_type_new (gtk_tree_item_pix_get_type ()));
  gtk_label_set (GTK_LABEL (GTK_TREE_ITEM_PIX (tree_item_pix)->label_widget),
    label);
  gtk_tree_item_pix_set_order (GTK_TREE_ITEM_PIX (tree_item_pix), location);
  GTK_TREE_ITEM_PIX (tree_item_pix)->collapsed_xpm = collapsed_xpm;
  if (expanded_xpm == NULL)
    GTK_TREE_ITEM_PIX (tree_item_pix)->expanded_xpm = collapsed_xpm;
  else
    GTK_TREE_ITEM_PIX (tree_item_pix)->expanded_xpm = expanded_xpm;

  return tree_item_pix;
}

void
gtk_tree_item_pix_set_order (GtkTreeItemPix *tree_item, guint location)
{
  tree_item->label_position = location;

  if (tree_item->table_widget)
  {
    gtk_container_remove (GTK_CONTAINER (tree_item->table_widget),
                           tree_item->label_widget);

    if (!GTK_TREE_ITEM (tree_item)->expanded)
      gtk_container_remove (GTK_CONTAINER (tree_item->table_widget),
                            tree_item->collapsed_pix_widget);
    else
      gtk_container_remove (GTK_CONTAINER (tree_item->table_widget),
                            tree_item->expanded_pix_widget);

    if (tree_item->label_position == GTK_TREE_ITEM_PIX_LABEL_LAST)
    {
      if (!GTK_TREE_ITEM (tree_item)->expanded)
        gtk_table_attach (GTK_TABLE (tree_item->table_widget),
                          tree_item->collapsed_pix_widget, 0, 1, 0, 1,
                          GTK_SHRINK, GTK_SHRINK, 0, 0);
      else
        gtk_table_attach (GTK_TABLE (tree_item->table_widget),
                          tree_item->expanded_pix_widget, 0, 1, 0, 1,
                          GTK_SHRINK, GTK_SHRINK, 0, 0);

      gtk_table_attach (GTK_TABLE (tree_item->table_widget),
                        tree_item->label_widget, 1, 2, 0, 1,
                        GTK_SHRINK, GTK_SHRINK, 0, 0);
    }
    else
    {
      if (!GTK_TREE_ITEM (tree_item)->expanded)
        gtk_table_attach (GTK_TABLE (tree_item->table_widget),
                          tree_item->collapsed_pix_widget, 1, 2, 0, 1,
                          GTK_SHRINK, GTK_SHRINK, 0, 0);
      else
        gtk_table_attach (GTK_TABLE (tree_item->table_widget),
                          tree_item->expanded_pix_widget, 1, 2, 0, 1,
                          GTK_SHRINK, GTK_SHRINK, 0, 0);

      gtk_table_attach (GTK_TABLE (tree_item->table_widget),
                        tree_item->label_widget, 0, 1, 0, 1,
                        GTK_SHRINK, GTK_SHRINK, 0, 0);
    }
  }
}

static void
tree_item_pix_collapse (GtkWidget *item, gpointer data)
{
  GtkTree* tree;
  GtkTreeItemPix* tree_item;

  g_return_if_fail (item != NULL);
  g_return_if_fail (GTK_IS_TREE_ITEM_PIX (item));

  if (GTK_TREE_ITEM (item)->subtree)
  {
    tree_item = GTK_TREE_ITEM_PIX (item);

    /* hide/show pixmaps */
    if (tree_item->table_widget)
    {
      gtk_container_remove (GTK_CONTAINER (tree_item->table_widget),
                            tree_item->expanded_pix_widget);
      gtk_container_remove (GTK_CONTAINER (tree_item->table_widget),
                            tree_item->label_widget);

      if (tree_item->label_position == GTK_TREE_ITEM_PIX_LABEL_LAST)
      {
        gtk_table_attach (GTK_TABLE (tree_item->table_widget),
                          tree_item->collapsed_pix_widget, 0, 1, 0, 1,
                          GTK_SHRINK, GTK_SHRINK, 0, 0);
        gtk_table_attach (GTK_TABLE (tree_item->table_widget),
                          tree_item->label_widget, 1, 2, 0, 1,
                          GTK_SHRINK, GTK_SHRINK, 0, 0);
      }
      else
      {
        gtk_table_attach (GTK_TABLE (tree_item->table_widget),
                          tree_item->collapsed_pix_widget, 1, 2, 0, 1,
                          GTK_SHRINK, GTK_SHRINK, 0, 0);
        gtk_table_attach (GTK_TABLE (tree_item->table_widget),
                          tree_item->label_widget, 0, 1, 0, 1,
                          GTK_SHRINK, GTK_SHRINK, 0, 0);
      }
    }
  }
}

static void
tree_item_pix_expand (GtkWidget *item, gpointer data)
{
  GtkTree* tree;
  GtkTreeItemPix* tree_item;

  g_return_if_fail (item != NULL);
  g_return_if_fail (GTK_IS_TREE_ITEM_PIX (item));

  if (GTK_TREE_ITEM (item)->subtree)
  {
    tree_item = GTK_TREE_ITEM_PIX (item);

    /* hide/show pixmaps */
    if (tree_item->table_widget)
    {
      gtk_container_remove (GTK_CONTAINER (tree_item->table_widget),
                            tree_item->collapsed_pix_widget);
      gtk_container_remove (GTK_CONTAINER (tree_item->table_widget),
                            tree_item->label_widget);

      if (tree_item->label_position == GTK_TREE_ITEM_PIX_LABEL_LAST)
      {
        gtk_table_attach (GTK_TABLE (tree_item->table_widget), 
                          tree_item->expanded_pix_widget, 0, 1, 0, 1,
                          GTK_SHRINK, GTK_SHRINK, 0, 0);
        gtk_table_attach (GTK_TABLE (tree_item->table_widget), 
                          tree_item->label_widget, 1, 2, 0, 1,
                          GTK_SHRINK, GTK_SHRINK, 0, 0);
      }
      else
      {
        gtk_table_attach (GTK_TABLE (tree_item->table_widget),
                          tree_item->expanded_pix_widget, 1, 2, 0, 1,
                          GTK_SHRINK, GTK_SHRINK, 0, 0);
        gtk_table_attach (GTK_TABLE (tree_item->table_widget),
                          tree_item->label_widget, 0, 1, 0, 1,
                          GTK_SHRINK, GTK_SHRINK, 0, 0);
      }
    }
  }
}

static void
gtk_tree_item_add_pixmaps (GtkTreeItemPix *tree_item)
{
  g_return_if_fail (tree_item != NULL);
  g_return_if_fail (GTK_IS_TREE_ITEM_PIX (tree_item));

  /* create pixmap for collapsed icon */
  tree_item->pixmap_collapsed =
    gdk_pixmap_create_from_xpm_d (GTK_WIDGET (tree_item)->window,
                                  &tree_item->pixmap_mask_collapsed,
                                  NULL,
                                  tree_item->collapsed_xpm);

  /* create pixmap for expanded icon */
  tree_item->pixmap_expanded =
    gdk_pixmap_create_from_xpm_d (GTK_WIDGET (tree_item)->window,
                                  &tree_item->pixmap_mask_expanded,
                                  NULL,
                                  tree_item->expanded_xpm);

  gdk_pixmap_ref (tree_item->pixmap_collapsed);
  gdk_bitmap_ref (tree_item->pixmap_mask_collapsed);
  gdk_pixmap_ref (tree_item->pixmap_expanded);
  gdk_bitmap_ref (tree_item->pixmap_mask_expanded);

  gtk_pixmap_set (GTK_PIXMAP (tree_item->collapsed_pix_widget),
                  tree_item->pixmap_collapsed,
                  tree_item->pixmap_mask_collapsed);
  gtk_pixmap_set (GTK_PIXMAP (tree_item->expanded_pix_widget),
                  tree_item->pixmap_expanded,
                  tree_item->pixmap_mask_expanded);
}

static void
gtk_tree_item_pix_remove_pixmaps (GtkTreeItemPix *tree_item)
{
  g_return_if_fail (tree_item != NULL);
  g_return_if_fail (GTK_IS_TREE_ITEM (tree_item));

  gdk_pixmap_unref (tree_item->pixmap_collapsed);
  gdk_bitmap_unref (tree_item->pixmap_mask_collapsed);
  gdk_pixmap_unref (tree_item->pixmap_expanded);
  gdk_bitmap_unref (tree_item->pixmap_mask_expanded);
}

static void
gtk_tree_item_pix_realize (GtkWidget *widget)
{
  g_return_if_fail (widget != NULL);
  g_return_if_fail (GTK_IS_TREE_ITEM_PIX (widget));

  if (GTK_WIDGET_CLASS (parent_class)->realize)
    (* GTK_WIDGET_CLASS (parent_class)->realize) (widget);

  gdk_window_set_background (widget->window,
                             &widget->style->base[GTK_STATE_NORMAL]);

  gtk_tree_item_add_pixmaps (GTK_TREE_ITEM_PIX (widget));
}

static void
gtk_tree_item_pix_destroy (GtkObject *object)
{
  GtkTreeItemPix* item;
  GtkWidget* child;

  g_return_if_fail (object != NULL);
  g_return_if_fail (GTK_IS_TREE_ITEM_PIX (object));

#ifdef TREE_DEBUG
  g_message("+ gtk_tree_pix_item_destroy [object %#x]\n", (int)object);
#endif /* TREE_DEBUG */

  item = GTK_TREE_ITEM_PIX (object);

  /* free sub tree if it exist */
  child = GTK_TREE_ITEM (item)->subtree;
  if (child)
    {
      gtk_widget_ref (child);
      gtk_widget_unparent (child);
      gtk_widget_destroy (child);
      gtk_widget_unref (child);
      GTK_TREE_ITEM (item)->subtree = NULL;
    }

  /* free table container */
  child = item->table_widget;
  if (child)
    {
      gtk_widget_ref (child);
      gtk_widget_unparent (child);
      gtk_widget_destroy (child);
      gtk_widget_unref (child);
      item->table_widget = NULL;
    }

  /* free lable widget */
  child = item->label_widget;
  if (child)
    {
      gtk_widget_ref (child);
      gtk_widget_unparent (child);
      gtk_widget_destroy (child);
      gtk_widget_unref (child);
      item->label_widget = NULL;
    }

  /* destroy collapsed pixmap */
  if (item->collapsed_pix_widget)
    {
      gtk_widget_destroy (item->collapsed_pix_widget);
      gtk_widget_unref (item->collapsed_pix_widget);
      item->collapsed_pix_widget = NULL;
    }

  /* destroy expanded pixmap */
  if (item->expanded_pix_widget)
    {
      gtk_widget_destroy (item->expanded_pix_widget);
      gtk_widget_unref (item->expanded_pix_widget);
      item->expanded_pix_widget = NULL;
    }

  /* By removing the pixmaps here, and not in unrealize, we depend on
   * the fact that a widget can never change colormap or visual.
   */
  gtk_tree_item_pix_remove_pixmaps (item);

  GTK_OBJECT_CLASS (parent_class)->destroy (object);

#ifdef TREE_DEBUG
  g_message("- gtk_tree_pix_item_destroy\n");
#endif /* TREE_DEBUG */
}
